/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file conn_mgr.h
 *
 * Defines the cconn_mgr class, which is a manager (and Abstract Factory) for
 * the cconn class. The cconn_mgr class provides methods to send a message to a
 * given user or address, to broadcast a message over all open connections, to
 * manage the connections (to actively discover or purge connections as
 * necessary), etc.
 *
 * The cconn_mgr class is implemented as a singleton. Use the instance method
 * to retrieve the cconn_mgr instance.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef _CONN_MGR_H
#define _CONN_MGR_H

#include "conn.h"      // have to so that i can declare a list<cconn> :(

#include <iostream>
#include <string>
#include <list>
#include <pthread.h>
using namespace std;


/*****************************************************************************
 * HELPER PROTOTYPES
 *****************************************************************************/
uint get_local_addr(struct sockaddr_in &sin);
struct in_addr resolve_addr(const char *addr);
void verify_socket(int sock);


/*****************************************************************************
 * FORWARD DECLARATIONS
 *****************************************************************************/
class cmsg;
struct p4_conn_info;  // for the client api functions


/*****************************************************************************
 * CCONN_MGR CLASS
 *****************************************************************************/
class cconn_mgr {
  public:

  /** instance
   *
   * Returns the only instance of cconn_mgr. This implements the singleton
   * pattern.
   */
  static cconn_mgr &instance();

  /** receive
   *
   * Receives data on any connection that has data pending, and accepts new
   * connections.
   *
   * NOTE: receive is *non-blocking* - it calls select() first with a minimal
   * timeout (10 msecs currently), and if there is no data pending on any open
   * socket, receive returns.
   */
  uint receive();


  /** manage
   *
   * Manages the currently open connections - if there are too few, sends a
   * DISCOVER message, and if there are too many, closes a connection.
   * Transient connections are closed preferentially before non-transient
   * connections, and connections that are currently receiving a message are
   * never closed.
   *
   * NOTE: Leaving transient connections open is an optimization. Right now,
   * we close them immediately after sending/receiving a message.
   */
  void manage();


  /** listen
   *
   * Start the listener socket, which listens for incoming P4 connections.
   * The port is given by the setting "p4_port".
   */
  uint listen();


  /** get_conn
   *
   * Searches the connection list for a connection to the given user. If it
   * exists, it is returned; otherwise, a failed connection is returned.
   */
  cconn &get_conn(const string &username);


  /** client_get_num_conns, client_get_conns
   *
   * These methods do the work for the p4_get_num_conns and p4_get_conns
   * client API functions. client_get_num_conns returns the number of
   * connections currently in progress, and client_get_conns returns
   * information about each of the connections.
   */
  void client_get_num_conns(int &num_conns) const;
  void client_get_conns(p4_conn_info *conn_arr, int num_conns,
                        int &num_written) const;

  /** num_connected
   *
   * Returns the number of connections in the connection list that are
   * actually connected.
   */
  uint num_connected() const;

  protected:
  friend class cmsg;          // so that messages can call send and broadcast
  friend class cconn;         // so that conns can call forward
  friend class cmsg_discover; // for random send
  friend class cmsg_connect;  // for make_conn_permanent


  /** send
   *
   * Sends the given message to the given user (or address). If a connection
   * to the user (or address) is already open, it is reused. Otherwise, a new
   * (transient) connection is opened. If the lock parameter is true, the
   * conn_mgr mutex will be acquired; otherwise, it won't be. This is an ugly
   * kludge, but necessary so that messages can send other messages in their
   * virt_process methods. (Such as DISCOVER, when you reply with a CONNECT
   * message instead of forwarding the DISCOVER.)
   *
   * The connection that was (or will be) used to send the message is
   * returned.
   *
   * NOTE: these send methods are *non-blocking* - if a connection needs
   * to be opened, send will return immediately. This may mean that the
   * message is sent later (after the call to send returns)!
   */
  const cconn &send(cmsg &msg, const string &username, bool lock = true);

  /** random_send
   *
   * Sends the msg along a random connection, but will not send along
   * a connection to the user except_user.  
   *
   * TODO: this function cannot block since it is called by manage()
   *          should it still be protected?
   */
  const cconn &random_send(cmsg &msg, const string &except_user);

  /** broadcast
   *
   * Sends the message over every open connection. If lock is true, the
   * conn_mgr lock is acquired; otherwise, it's not. (This is necessary so
   * that broadcast messages can forward themselves from within process.)
   *
   * Note that this is called both to broadcast messages that originated at
   * this client, and to forward incoming messages that have the broadcast
   * flag set. (Since the username field is the original sender, not the
   * username of the last hop, we wouldn't know which connection to avoid
   * sending on anyway.)
   */
  uint broadcast(cmsg &msg, bool lock = true);

  /** make_conn_permanent 
   *
   * Changes the status of the conn that received msg to 
   * permanent.
   */
  void make_conn_permanent(const string &username);

  /** random_conn
   *
   * Returns an iterator to a random connection in the connection list.
   */
  list<cconn>::iterator cconn_mgr::random_conn();
  
  /** ctor
  */
  cconn_mgr();

  /** ctor
  */
  ~cconn_mgr();


  /**************************************************************************
   * CONSTANTS AND MEMBERS
   **************************************************************************/
  public:
  static const uint kideal_min_conns, kideal_max_conns;
  static const uint kselect_timeout_msecs;
  static const float kdiscover_prob;
  static const ushort kp4_port;

  private:
  list<cconn> conns;
  int listener;       // listener socket
  pthread_mutex_t mtx;
};


/*****************************************************************************
 * INLINE METHODS
 *****************************************************************************/
inline cconn_mgr &cconn_mgr::instance()
{
  static cconn_mgr conn_mgr;

  return conn_mgr;
}


#endif  // _CONN_MGR_H

